class Unit:

    def __init__(self, factor, m=0, kg=0, s=0, amp=0, cd=0, mol=0, k=0):
        """

        Args:
            factor:
            m:
            kg:
            s:
            amp:
            cd:
            mol:
            k:
        """
        self._factor = factor
        self._units = {
            'm'   : m,
            'kg'  : kg,
            's'   : s,
            'k'   : k,
            'amp' : amp,
            'cd'  : cd,
            'mol' : mol
        }

    def normalize(self, quantity):
        return quantity * self._factor

    def from_base_units(self, normalized_quantity):
        return normalized_quantity / self._factor

    def convert(self, reference_unit):
        assert(self.matches_dimensions(reference_unit))
        return reference_unit.from_base_units(self._factor)

    def _exponent(self, unit_key):
        """

        Args:
            unit_key:

        Returns:

        """
        return self._units[unit_key]

    def __mul__(self, other):
        """

        Args:
            other: Unit:

        Returns:

        """
        new_units = {
            key: value + other._units[key]
            for key, value in self._units.items()
        }
        return Unit(self._factor * other._factor, **new_units)

    __rmul__ = __mul__

    def __truediv__(self, other):
        """

        Args:
            other:

        Returns:

        """
        new_units = {
            key: value - other._units[key]
            for key, value in self._units.items()
        }
        return Unit(self._factor / other._factor, **new_units)

    def __pow__(self, exp):
        """

        Args:
            exp:

        Returns:

        """
        new_units = {
            key: value * exp
            for key, value in self._units.items()
        }
        return Unit(self._factor ** exp, **new_units)

    def dimensions(self):
        return '*'.join(
            '%s^%i' % (unit, exp)
            for (unit, exp)
            in self._units.items()
            if exp != 0
        )

    def __eq__(self, other):
        return self._factor == other._factor and self.matches_dimensions(other)

    def matches_dimensions(self, other):
        return all(self._units[key] == other._units[key] for key in self._units)

    def __str__(self):
        return '%f %s' % (self._factor, self.dimensions())

    def __repr__(self):
        return 'Unit(%f, m=%i, kg=%i, s=%i, k=%i, amp=%i, cd=%i, mol=%i)' % (
            self._factor,
            self._units['m'],
            self._units['kg'],
            self._units['s'],
            self._units['k'],
            self._units['amp'],
            self._units['cd'],
            self._units['mol']
        )

# SI Convenience (unitless)

nano = Unit(1 / 1e9)
micro = Unit(1 / 1e6)
milli = Unit(1 / 1e3)
centi = Unit(1 / 1e2)
deci = Unit(1 / 10)
unity = Unit(1)
deca = Unit(10)
hecta = Unit(1e2)
kilo = Unit(1e3)
mega = Unit(1e6)
giga = Unit(1e9)

# base units

meters = Unit(1, m=1)
kilograms = Unit(1, kg=1)
seconds = Unit(1, s=1)
candelas = Unit(1, cd=1)
mols = Unit(1, mol=1)
amps = Unit(1, amp=1)
kelvin = Unit(1, k=1)

# Distance

centimeters = centi * meters
millimeters = milli * meters
inches = Unit(25.4) * millimeters
feet = Unit(12) * inches
yards = Unit(3) * feet
miles = Unit(5280)

square_meters = meters ** 2
square_centimeters = centimeters ** 2
square_millimeters = millimeters ** 2
square_inches = inches ** 2
square_feet = feet ** 2

cubic_meters = meters ** 3
cubic_centimeters = centimeters ** 3
cubic_millimeters = millimeters ** 3

quartic_meters = meters ** 4
quartic_centimeters = centimeters ** 4
quartic_millimeters = millimeters ** 4

# Mass

grams = kilograms / kilo
tonnes = Unit(1000) * kilograms
pounds = Unit(0.45359237) * kilograms
slugs = Unit(14.5939) * kilograms

# Time

minutes = Unit(60) * seconds
hours = Unit(60) * minutes

# Velocity

meters_per_second = meters / seconds
miles_per_hour = miles / hours

# Force

newtons = kilograms * meters / seconds ** 2
kilonewtons = kilo * newtons
dynes = grams * centimeters / seconds ** 2
pound_force = slugs * feet / seconds ** 2

# Pressure / Stress

pascals = newtons / meters ** 2
kilopascals = kilo * pascals
megapascals = mega * pascals
gigapascals = giga * pascals

atmospheres = Unit(101325) * pascals
torrs = atmospheres / Unit(760)


joules = Unit(1, kg=1, m=2, s=-2)
watts = joules / seconds

volts = watts / amps

newton_meters = newtons * meters
kilonewton_meters = kilonewtons * meters
pound_feet = pound_force * feet

# Angles

PI = 3.14159265358979323846264338327950288419716939937510582097

radians = Unit(1)
degrees = radians * Unit(PI / 180)
arcminutes = degrees / Unit(60)
arcseconds = arcminutes / Unit(60)


hertz = Unit(1) / seconds
rpm = Unit(1) / minutes




"""

"""

from .meta import Quantity, LinearQuantity

import siquant.units as u

class Scalar(metaclass=LinearQuantity, base_units=u.unity):
    pass

class Distance(metaclass=LinearQuantity, base_units=u.meters):

    def squared(self):
        return Area(self.get() ** 2)

    def cubed(self):
        return Volume(self.get() ** 3)

class Area(metaclass=LinearQuantity, base_units=u.square_meters):
    pass

class Volume(metaclass=LinearQuantity, base_units=u.cubic_meters):
    pass

class Time(metaclass=LinearQuantity, base_units=u.seconds):
    pass

class Velocity(metaclass=LinearQuantity, base_units=u.meters / u.seconds):
    pass

class Acceleration(metaclass=LinearQuantity, base_units=u.meters / u.seconds ** 2):
    pass

Acceleration.G = Acceleration(9.80665)

class Mass(metaclass=LinearQuantity, base_units=u.kilograms):
    pass

class Work(metaclass=LinearQuantity, base_units=u.joules):
    pass

class Power(metaclass=LinearQuantity, base_units=u.watts):
    pass

class Density(metaclass=LinearQuantity, base_units=u.kilograms / u.cubic_meters):
    pass

class Force(metaclass=LinearQuantity, base_units=u.newtons):
    pass

class Moment(metaclass=LinearQuantity, base_units=u.newton_meters):
    pass

class Stress(metaclass=LinearQuantity, base_units=u.pascals):
    pass

class MomentOfArea(metaclass=LinearQuantity, base_units=u.quartic_meters):
    pass

class SectionModulus(metaclass=LinearQuantity, base_units=u.cubic_meters):
    pass

class Angle(metaclass=LinearQuantity, base_units=u.radians):
    pass

class Frequency(metaclass=Quantity, base_units=u.hertz):
    pass

#todo: TrigAngle with cached trig functions?



def bin_op_dispatcher():
    registry = {}
    def dispatch(inst, other):
        fcn = registry[type(other)]
        return fcn(inst, other)
    dispatch.supported_types = lambda: registry.keys()

    def register(*op_types):
        def decorator(fcn):
            for op_type in op_types:
                registry[op_type] = fcn
            return dispatch
        return decorator

    return (dispatch, register)


from functools import total_ordering

from .dispatch import bin_op_dispatcher

def equals(q1, q2):
    return type(q1) == type(q2) and q1._quantity == q2._quantity

def less_than(q1, q2):
    return type(q1) == type(q2) and q1._quantity < q2._quantity

class Quantity(type):

    def __new__(meta, name, bases, attrs, base_units=None):
        if base_units is None:
            raise ValueError('base_units is required to define a new quantity type')

        defined_slots = attrs.get('__slots__', ())

        def __init__(self, quantity, units=base_units):
            assert(base_units.matches_dimensions(units))
            self._quantity = units.normalize(quantity)

        def get_as(self, units):
            assert(base_units.matches_dimensions(units))
            return units.from_base_units(self._quantity)

        attrs = {
            **attrs,

            'base_units' : base_units,

            '__slots__' : ( '_quantity', *defined_slots),

            '__init__' : __init__,

            'get' : lambda self: self._quantity,

            'get_as' : get_as,

            '__eq__' : equals,

            '__lt__' : less_than,

            '__neg__' : lambda self: self.__class__(-self._quantity),

            '__str__' : lambda self: '%f %s' % (self._quantity, base_units.dimensions()),

            '__repr__' : lambda self: '%s(%f, units=%r)' % (self.__class__.__name__, self._quantity, base_units),

        }

        QType = super().__new__(meta, name, bases, attrs)

        return total_ordering(QType)

def add(q1, q2):
    assert(type(q1) == type(q2))
    return q1.__class__(q1._quantity + q2._quantity)

def sub(q1, q2):
    assert(type(q1) == type(q2))
    return q1.__class__(q1._quantity - q2._quantity)

class LinearQuantity(Quantity):

    def __new__(meta, name, bases, attrs, base_units=None):
        QType = super().__new__(meta, name, bases, attrs, base_units=base_units)

        (QType.__add__, QType.adder) = bin_op_dispatcher()
        (QType.__sub__, QType.subtracter) = bin_op_dispatcher()
        (QType.__mul__, QType.multiplier) = bin_op_dispatcher()
        (QType.__truediv__, QType.divider) = bin_op_dispatcher()

        QType.__rmul__ = QType.__mul__

        QType.multiplier(float, int)(lambda self, scalar: QType(self.get() * scalar))
        QType.divider(float, int)(lambda self, scalar: QType(self.get() / scalar))
        QType.divider(QType)(lambda self, other: self.get() / other.get())
        QType.adder(QType)(add)
        QType.subtracter(QType)(sub)

        def q_multiplier(op_type, result_type):
            def mul(q1, q2):
                return result_type(q1._quantity * q2._quantity, q1.base_units * q2.base_units)
            return QType.multiplier(op_type)(mul)
        QType.q_multiplier = q_multiplier

        def q_divider(op_type, result_type):
            def div(q1, q2):
                return result_type(q1._quantity / q2._quantity, q1.base_units / q2.base_units)
            return QType.divider(op_type)(div)
        QType.q_divider = q_divider

        return QType




import unittest

import siquant.units as u

class BaseUnitTestCase(unittest.TestCase):

    def test_base_units_at_unity(self):

        self.assertEqual(u.meters._factor, 1)
        self.assertEqual(u.square_meters._factor, 1)
        self.assertEqual(u.cubic_meters._factor, 1)
        self.assertEqual(u.quartic_meters._factor, 1)
        self.assertEqual(u.kilograms._factor, 1)
        self.assertEqual(u.seconds._factor, 1)
        self.assertEqual(u.newtons._factor, 1)
        self.assertEqual(u.newton_meters._factor, 1)
        self.assertEqual(u.pascals._factor, 1)
        self.assertEqual(u.hertz._factor, 1)
        self.assertEqual(u.radians._factor, 1)

    def test_distance_units(self):

        self.assertTrue(u.meters.matches_dimensions(u.meters))
        self.assertTrue(u.millimeters.matches_dimensions(u.meters))
        self.assertTrue(u.centimeters.matches_dimensions(u.meters))

        self.assertFalse(u.square_meters.matches_dimensions(u.meters))
        self.assertFalse(u.cubic_meters.matches_dimensions(u.meters))
        self.assertFalse(u.quartic_meters.matches_dimensions(u.meters))

        self.assertFalse(u.seconds.matches_dimensions(u.meters))
        self.assertFalse(u.kilograms.matches_dimensions(u.meters))

        self.assertFalse(u.newtons.matches_dimensions(u.meters))
        self.assertFalse(u.newton_meters.matches_dimensions(u.meters))

        self.assertEqual(u.meters.from_base_units(100), 100)
        self.assertEqual(u.millimeters.from_base_units(100), 100000)
        self.assertEqual(u.centimeters.from_base_units(100), 10000)

    def test_area_units(self):

        self.assertTrue(u.square_meters.matches_dimensions(u.square_meters))
        self.assertTrue(u.square_centimeters.matches_dimensions(u.square_meters))
        self.assertTrue(u.square_millimeters.matches_dimensions(u.square_meters))






if __name__ == '__main__':
    unittest.main()
